#include <stdio.h> // ###
#include "lpc210x.h"
#include "mmc.h"
#include "spi.h"

/*************************************************************************
 *
 *    Used with ICCARM and AARM.
 *
 *    (c) Copyright IAR Systems 2003
 *
 *    File name   : mmc_ll.h
 *    Description : define MMC module
 *
 *    History :
 *    1. Data        : July 1, 2005
 *       Author      : Stanimir Bonev
 *       Description : Create
 *    2. Data        : July 1, 2005
 *       Author      : Stanimir Bonev
 *       Description : Modify
 *        Fix a lock problem in spiReceiveBlock
 *
 *    $Revision: 1.4 $
 * (C) Joel Winarske, 2006,2007                                        
**************************************************************************/

/*************************************************************************
 * Function Name: spiChipSelect
 * Parameters: Boolean Select
 * Return: none
 *
 * Description: MMC Chip select control
 * Select = true  - Chip is enabled
 * Select = false - Chip is disabled
 *
 *************************************************************************/
void spiChipSelect (BOOL select)
{
  if (select)
    GPIO0_IOCLR = GPIO_IO_P20;
  else
  {
    GPIO0_IOSET = GPIO_IO_P20;
    
    while (!(SSP_SR & SSP_SR_TNF))
      ;

    SSP_DR = 0xff;

    //
    // Wait until TX fifo and TX shift buffer are empty
    //
    while (SSP_SR & SSP_SR_BSY)
      ;
    while (!(SSP_SR & SSP_SR_RNE))
      ;

    do
    {
      select = SSP_DR;
    }
    while (SSP_SR & SSP_SR_RNE);
  }
}

/*************************************************************************
 * Function Name: spiPresent
 * Parameters: none
 * Return: Boolean - true card present
 *                 - false card not present
 *
 * Description: MMC present check
 *
 *************************************************************************/
inline BOOL spiPresent (void)
{
  return TRUE;
}

/*************************************************************************
 * Function Name: spiWriteProtect
 * Parameters: none
 * Return: Boolean - true card is protected
 *                 - false card not protected
 *
 * Description: MMC Write protect check
 *
 *************************************************************************/
inline BOOL spiWriteProtect (void)
{
  return FALSE;
}

/*************************************************************************
 * Function Name: spiSetClockFreq
 * Parameters: Int32U Frequency
 * Return: Int32U
 *
 * Description: Set SPI ckl frequency
 *
 *************************************************************************/
U32 spiSetClockFreq (U32 frequency)
{
  U32 Div = SystemPeripherialClock / frequency;

  if (Div < 2)
    Div = 2;
  else if (Div > 254)
    Div = 254;

  ++Div; 
  Div &= ~1;
  SSP_CPSR = Div;

  return (SystemPeripherialClock / Div);
}

/*************************************************************************
 * Function Name: spiInit
 * Parameters: none
 * Return: int
 * 0 - no error
 * 1 - speed is to high
 *
 * Description: Init SPI, Cart Present, Write Protect and Chip select pins
 *
 *************************************************************************/
void spiInit (void)
{
  U32 i;
  volatile U32 dummy;

  //
  //  Chip select
  //
  GPIO0_IOSET = GPIO_IO_P20;
  GPIO0_IODIR |= GPIO_IO_P20;

  //
  //  SPI init
  //
  SCB_PCONP |= SCB_PCONP_PCSPI1;

  SSP_CR0  = SSP_CR0_DSS_8 | SSP_CR0_FRF_SPI;
  SSP_CR1  = 0x00;
  SSP_IMSC = 0x00;

  //
  //  Clock Freq. Identification Mode < 400kHz
  //
  spiSetClockFreq (400000);

  PCB_PINSEL1 |=  PCB_PINSEL1_P017_SCK1;
  PCB_PINSEL1 |=  PCB_PINSEL1_P018_MISO1;
  PCB_PINSEL1 |=  PCB_PINSEL1_P019_MOSI1;
  PCB_PINSEL1 &= ~PCB_PINSEL1_P020_MASK;

  //
  //  Enable SPI
  //
  SSP_CR1 |= SSP_CR1_SSE;

  for (i = 0; i < 8; i++)
    dummy = SSP_DR;
}

/*************************************************************************
 * Function Name: spiTransferByte
 * Parameters: U8 ch
 * Return: U8
 *
 * Description: Read byte from SPI
 *
 *************************************************************************/
U8 spiTransferByte (U8 c)
{
  while (!(SSP_SR & SSP_SR_TNF))
    ;

  SSP_DR = c;

  while (!(SSP_SR & SSP_SR_RNE))
    ;

  return SSP_DR;
}

/*-----------------------------------------------------------------------*/
/* Wait for card ready                                                   */
/*-----------------------------------------------------------------------*/
U8 spiWaitReady (void)
{
  U8 res;

  do                        /* TODO: add in 500ms timeout */
  {      
    SSP_DR = 0xff;

    while (!(SSP_SR & SSP_SR_RNE))
      ;

    res = SSP_DR;
  }
  while (res != 0xff);

  return res;
}

/*************************************************************************
 * Function Name: spiSendBlock
 * Parameters: pInt8U pData, Int32U Size
 *
 * Return: void
 *
 * Description: Read byte from SPI
 *
 *************************************************************************/
void spiSendBlock (U8 *pData, U32 size)
{
  U32 outCount = size;
  volatile U32 dummy;

  while (outCount)
  {
    while ((SSP_SR & SSP_SR_TNF) && outCount)
    {
      SSP_DR = *pData++;
      --outCount;
    }
  }

  while ((SSP_SR & SSP_SR_RNE) || !(SSP_SR & SSP_SR_TFE))
    dummy = SSP_DR;
}

/*************************************************************************
 * Function Name: spiReceiveBlock
 * Parameters: pInt8U pData, Int32U Size
 *
 * Return: void
 *
 * Description: Read byte from SPI
 *
 *************************************************************************/
void spiReceiveBlock (U8 *pData, U32 size)
{
  U32 delta = 0;

  while (size || delta)
  {
    while ((SSP_SR & SSP_SR_TNF) && (delta < SSP_FIFO_DEPTH) && size)
    {
      SSP_DR = 0xff;
      --size; 
      ++delta;
    }

    while (SSP_SR & SSP_SR_RNE)
    {
      *pData++ = SSP_DR;
      --delta;
    }
  }
}

/*************************************************************************
 * Function Name: spiDly_1ms
 * Parameters: Int32U Delay
 * Return: none
 *
 * Description: Delay [msec]
 *
 *************************************************************************/
void spiDelay1ms (U32 delay)
{
  volatile U32 i;

  for (;delay; --delay)
  {
    for (i = MMC_DLY_1MSEC; i; --i)
      ;
  }
}
